
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

extern void	Error();

#define LEN	1023

typedef struct List {
		char		*mask,*data;
		double		base,step;
		int		Dx,Dy,Nx,Ny;
		struct List	*next;
	} List;

typedef struct {
		int	i;
		double	d;
	} Sum;

#define ALLOC(type,size)	(type *)malloc((size_t)(size)*sizeof(type))

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	mask[LEN],data[LEN],desc[LEN],
		*bis=" : bad image size.",
		*mae="memory allocation error.";
	int	Dx,Dy,mLen,dLen,x,y,x1,y1,x2,y2,X1,Y1,X2,Y2,BPS;
	double	base,step,roof,none,val,min,max;
	List	*list,*top=NULL,**ptr=(&top);
	Sum	**sum,**sy,*sx;
	Cell	**out,**mIn,**dIn;

	if (argc!=4 && argc!=6 && argc!=8 && argc!=10)
	    Error("usage : add_2d_wm none {base step} BPS {x1 y1 x2 y2} TIFF");

	while (scanf("%s %s %lf %lf %d %d",mask,data,&base,&step,&Dx,&Dy)==6) {
	    for (mLen=0; mask[mLen]!='\0'; mLen++) ;
	    for (dLen=0; data[dLen]!='\0'; dLen++) ;

	    if ((list=ALLOC(List,1))==NULL ||
		(list->mask=ALLOC(char,mLen+1))==NULL ||
		(list->data=ALLOC(char,dLen+1))==NULL) Error(mae);

	    do list->mask[mLen]=mask[mLen]; while (--mLen>=0);
	    do list->data[dLen]=data[dLen]; while (--dLen>=0);

	    list->base=base;
	    list->step=step;

	    ReadImageFile(mask,&(list->Nx),&(list->Ny),NULL,NULL,NULL);
	    ReadImageFile(data,&x,&y,NULL,NULL,NULL);
	    if (list->Nx!=x || list->Ny!=y) {
		(void)fputs(data,stderr); Error(bis);
	    }
	    x2=(x1=list->Dx=Dx)+x-1;
	    y2=(y1=list->Dy=Dy)+y-1;
	    if (top==NULL) {
		X1=x1; X2=x2;
		Y1=y1; Y2=y2;
	    }
	    else {
		if (x1<X1) X1=x1; if (x2>X2) X2=x2;
		if (y1<Y1) Y1=y1; if (y2>Y2) Y2=y2;
	    }
	    *ptr=list; *(ptr=(&(list->next)))=NULL;
	}
	if (top==NULL) Error("no image.");

	if (argc==8 || argc==10) {
	    if ((x1=atoi(argv[argc-5]))>
		(x2=atoi(argv[argc-3])) || x1>X2 || x2<X1 ||
		(y1=atoi(argv[argc-4]))>
		(y2=atoi(argv[argc-2])) || y1>Y2 || y2<Y1) Error("bad area.");

	    X1=x1; Y1=y1; X2=x2; Y2=y2;
	}
	if ((BPS=atoi(argv[(argc==4 || argc==6)?argc-2:argc-6]))<1 ||
	    BPS>8*sizeof(Cell)) Error("bad BPS.");

	if (argc==6 || argc==10) {
	    if ((step=atof(argv[3]))<=0.0) Error("bad step.");

	    base=atof(argv[2]);
	    roof=base+step*(double)((1<<BPS)-1);
	}
	none=atof(argv[1]);

	X2-=(X1-1);
	Y2-=(Y1-1);

	if ((out=(Cell **)(sum=sy=ALLOC(Sum *,Y2)))==NULL) Error(mae);

	for (y=0; y<Y2; y++, sy++) {
	    if ((*sy=sx=ALLOC(Sum,X2))==NULL) Error(mae);

	    for (x=0; x<X2; x++, sx++) sx->d=(double)(sx->i=0);
	}
	for (list=top; list!=NULL; list=list->next) {
	    ReadImageFile(list->mask,&x,&y,NULL,&mIn,NULL);
	    if (x!=list->Nx || y!=list->Ny) {
		(void)fputs(list->mask,stderr); Error(bis);
	    }
	    ReadImageFile(list->data,&x,&y,NULL,&dIn,NULL);
	    if (x!=list->Nx || y!=list->Ny) {
		(void)fputs(list->data,stderr); Error(bis);
	    }
	    Dx=list->Dx-X1; x1=(Dx>=0)?0:-Dx; if ((x2=X2-Dx)>x) x2=x;
	    Dy=list->Dy-Y1; y1=(Dy>=0)?0:-Dy; if ((y2=Y2-Dy)>y) y2=y;
	    for (sy=sum  +Dy+(y=y1); y<y2; y++, sy++)
	    for (sx=(*sy)+Dx+(x=x1); x<x2; x++, sx++)
		if (mIn[y][x]) {
		    ++sx->i; sx->d+=(list->base+list->step*(double)dIn[y][x]);
		}

	    for (y=list->Ny-1; y>=0; y--) free(dIn[y]); free(dIn);
	    for (y=list->Ny-1; y>=0; y--) free(mIn[y]); free(mIn);
	}
	for (sy=sum,   y=0; y<Y2; y++, sy++)
	for (sx=(*sy), x=0; x<X2; x++, sx++) {
	    val=(sx->i)?sx->d/(double)sx->i:none;

	    if (x==0 && y==0)
		min=max=val;
	    else
		if (val<min) min=val; else if (val>max) max=val;
	}
	if (argc==4 || argc==8) {
	    if (min==max) (void)fputs("no value range (warning).\n",stderr);

	    step=(max-min)/(double)((1<<BPS)-1); base=min; roof=max;
	}
	(void)sprintf(desc,"%lf\t%lf\t%lf\t%18.16lf",min,max,base,step);
	(void)puts(desc);

	for (sy=sum,   y=0; y<Y2; y++, sy++)
	for (sx=(*sy), x=0; x<X2; x++, sx++) {
	    val=(sx->i)?sx->d/(double)sx->i:none;

	    out[y][x]=(val<=base)?0:
		      (val>=roof)?(Cell)((1<<BPS)-1):
				  (Cell)((val-base)/step+0.5);
	}
	StoreImageFile(argv[argc-1],X2,Y2,BPS,out,desc);

	return 0;
}
