
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "FFT.h"

extern void	Error();

static void	Trim(Ng,Nh,D1,D2,Og,Oh,D0)
int		*Ng,*Nh,*D1,*D2,*Og,*Oh,*D0;
{
	int	ng,nh;

	if ((ng=(*Nh)+(*D2))>(*Ng)) ng=(*Ng);
	if ((nh=(*Ng)-(*D1))>(*Nh)) nh=(*Nh);

	if (*D2<0) {
	    *Og=0; *Oh=(-((*D0)=(*D2)));
	}
	else if (*D1>0) {
	    *Og=(*D0)=(*D1); *Oh=0;
	}
	else
	    *Og=(*Oh)=(*D0)=0;

	*Ng=ng-(*Og);
	*Nh=nh-(*Oh); *D1-=(*D0); *D2-=(*D0);
}

#define BYTES(type,size)	(sizeof(type)*(size_t)(size))

static Complex	*Expand(L,E)
int		L;
Complex		*E;
{
	Complex	*e;
	int	l;

	if ((E=(Complex *)realloc(E,BYTES(Complex,L*2)))!=NULL)
	    for (e=E, l=0; l<L; l++, e++) {
		e[L].r=(-e->r);
		e[L].i=(-e->i);
	    }

	return E;
}

static Complex	Product(r1,i1,r2,i2)
double		r1,i1,r2,i2;
{
	Complex	c;

	c.r=r1*r2-i1*i2;
	c.i=r1*i2+i1*r2; return c;
}

#define ALLOC(type,size)	(type *)malloc(BYTES(type,size))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Ngx,Ngy,Nhx,Nhy,Dx1,Dy1,Dx2,Dy2,xg,yg,xh,yh,Dx0,Dy0,Lx,Ly,y,x,
		Lx1,Ly1,Lx2,Ly2,Ngx1,Ngy1,Nhx1,Nhy1,gy,hy,gx,hx,Lxx,Nghx,Nghy;
	Cell	**G,**H,**Gy,**Hy,*Gx,*Hx;
	double	Bg,Sg,Bh,Sh,**P,**Py,f,Ngx2,Ngy2,Nhx2,Nhy2,
		t,r,i,w,rg,ig,rh,ih,*Px,L,LNy,msd;
	Complex	*Ex,*Ey,*U,*V,**Q,**Qy,*Qx,*u,*ey,*v,g2,h2,*ex,c;
	char	*mae="memory allocation error.";

	if (argc!=7 && argc!=11)
Error("usage : rmsd_2d TIFF1 base1 step1 TIFF2 base2 step2 {Dx1 Dy1 Dx2 Dy2}");

	ReadImageFile(argv[1],&Ngx,&Ngy,NULL,&G,NULL);
	Bg=atof(argv[2]);
	Sg=atof(argv[3]);

	ReadImageFile(argv[4],&Nhx,&Nhy,NULL,&H,NULL);
	Bh=atof(argv[5]);
	Sh=atof(argv[6]);

	if (argc==7) {
	    Dx1=1-Nhx; Dx2=Ngx-1;
	    Dy1=1-Nhy; Dy2=Ngy-1;
	}
	else
	    if ((Dx1=atoi(argv[ 7]))>
		(Dx2=atoi(argv[ 9])) || Dx1<=(-Nhx) || Dx2>=Ngx ||
		(Dy1=atoi(argv[ 8]))>
		(Dy2=atoi(argv[10])) || Dy1<=(-Nhy) || Dy2>=Ngy)
		Error("bad range of distance.");

	Trim(&Ngx,&Nhx,&Dx1,&Dx2,&xg,&xh,&Dx0);
	Trim(&Ngy,&Nhy,&Dy1,&Dy2,&yg,&yh,&Dy0);

#define SETUPFFT(Ng,Nh,D1,D2,E)	\
	SetUpFFT(((Ng>Nh)?Ng:Nh)+ \
		 ((D2<0)?1-D1:(D1>0)?D2+1:(-D1>D2)?1-D1:D2+1),&E)

	if ((Lx=SETUPFFT(Ngx,Nhx,Dx1,Dx2,Ex))==0 ||
	    (Ex=Expand(Lx,Ex))==NULL ||
	    (Ly=SETUPFFT(Ngy,Nhy,Dy1,Dy2,Ey))==0 ||
	    (Ey=Expand(Ly,Ey))==NULL ||
	    (U=ALLOC(Complex,Ly))==NULL ||
	    (V=ALLOC(Complex,Ly))==NULL ||
	    (P=Py=ALLOC(double *,Dy2-Dy1+1))==NULL ||
	    (Q=Qy=ALLOC(Complex *,Ly))==NULL) Error(mae);

	for (y=Dy1; y<=Dy2; y++)
	    if ((*(Py++)=ALLOC(double,Dx2-Dx1+1))==NULL) Error(mae);

	for (Gy=G+yg, Hy=H+yh, y=0; y<Ly; y++, Qy++) {
	    if ((*Qy=ALLOC(Complex,Lx))==NULL) Error(mae);

	    if (y<Ngy)
		for (Qx=(*Qy), Gx=(*(Gy++))+xg, x=0; x<Lx; x++, Qx++)
		    if (x<Ngx) {
			f=Bg+Sg*(double)(*(Gx++)); Qx->r=f*f;
		    }
		    else
			Qx->r=0.0;
	    else
		for (Qx=(*Qy), x=0; x<Lx; x++, Qx++) Qx->r=0.0;

	    if (y<Nhy)
		for (Qx=(*Qy), Hx=(*(Hy++))+xh, x=0; x<Lx; x++, Qx++)
		    if (x<Nhx) {
			f=Bh+Sh*(double)(*(Hx++)); Qx->i=f*f;
		    }
		    else
			Qx->i=0.0;
	    else
		for (Qx=(*Qy), x=0; x<Lx; x++, Qx++) Qx->i=0.0;

	    if (y<Ngy || y<Nhy) FFT(Lx,Ex,*Qy,-1);
	}
	Lx1=Lx-1; Ly1=Ly-1;

	Lx2=Lx*2; Ly2=Ly*2;
	Ngx1=Ngx-1; Ngy1=Ngy-1;
	Nhx1=Nhx-1; Nhy1=Nhy-1;
	Ngx2=(double)(Ngx%2); Ngy2=(double)(Ngy%2);
	Nhx2=(double)(Nhx%2); Nhy2=(double)(Nhy%2);
/*
    x=0
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][0];

	FFT(Ly,Ey,U,-1);

	U->r=U->r*(double)Nhx*(double)Nhy+(double)Ngx*(double)Ngy*U->i;
	U->i=0.0;
	for (ey=Ey+1, gy=Ngy1, hy=Nhy1, u=U+1, v=U+Ly1;
	     u<v;
	     ey++, gy=(gy+Ngy1)%Ly2, hy=(hy+Nhy1)%Ly2, u++, v--) {
	    t=ey->r/ey->i;
	    r=Ey[hy].r;
	    i=Ey[hy].i; w=(double)Nhx*(r+i*t);
			g2=Product(u->r+v->r,u->i-v->i,w*r, w*i);
	    r=Ey[gy].r;
	    i=Ey[gy].i; w=(double)Ngx*(r+i*t);
			h2=Product(w*r,-w*i,u->i+v->i,u->r-v->r);

	    v->r=u->r=(g2.r+h2.r)/2.0;
		 u->i=(g2.i+h2.i)/2.0; v->i=(-u->i);
	}
	u->r=u->r*(double)Nhx*Nhy2+(double)Ngx*Ngy2*u->i;
	u->i=0.0;

	FFT(Ly,Ey,U, 1);

	for (u=U, y=0; y<Ly; y++) Q[y][0]=(*(u++));
/*
    x=1~Lx/2-1
*/
	for (ex=Ex+1, gx=Ngx1, hx=Nhx1, x=1, Lxx=Lx1;
	     x<Lxx;
	     ex++, gx=(gx+Ngx1)%Lx2, hx=(hx+Nhx1)%Lx2, x++, Lxx--) {
	    t=ex->r/ex->i;
	    r=Ex[hx].r;
	    i=Ex[hx].i; w=r+i*t; rh=w*r; ih=w*i;
	    r=Ex[gx].r;
	    i=Ex[gx].i; w=r+i*t; rg=w*r; ig=w*i;

	    for (u=U, v=V, y=0; y<Ly; y++) {
		*(u++)=Q[y][  x];
		*(v++)=Q[y][Lxx];
	    }
	    FFT(Ly,Ey,U,-1);
	    FFT(Ly,Ey,V,-1);

	    g2=Product(U->r+V->r,U->i-V->i,rh*(double)Nhy, ih*(double)Nhy);
	    h2=Product(rg*(double)Ngy,-ig*(double)Ngy,U->i+V->i,U->r-V->r);

	    V->r=U->r=(g2.r+h2.r)/2.0;
		 U->i=(g2.i+h2.i)/2.0; V->i=(-U->i);
	    for (ey=Ey+1, gy=Ngy1, hy=Nhy1, u=U+1, v=V+Ly1, y=1;
		 y<Ly;
		 ey++, gy=(gy+Ngy1)%Ly2, hy=(hy+Nhy1)%Ly2, u++, v--, y++) {
		t=ey->r/ey->i;
		r=Ey[hy].r;
		i=Ey[hy].i; w=r+i*t; c=Product(rh,ih,w*r,w*i);
				     g2=Product(u->r+v->r,u->i-v->i,c.r, c.i);
		r=Ey[gy].r;
		i=Ey[gy].i; w=r+i*t; c=Product(rg,ig,w*r,w*i);
				     h2=Product(c.r,-c.i,u->i+v->i,u->r-v->r);

		v->r=u->r=(g2.r+h2.r)/2.0;
		     u->i=(g2.i+h2.i)/2.0; v->i=(-u->i);
	    }
	    FFT(Ly,Ey,U, 1);
	    FFT(Ly,Ey,V, 1);

	    for (u=U, v=V, y=0; y<Ly; y++) {
		Q[y][  x]=(*(u++));
		Q[y][Lxx]=(*(v++));
	    }
	}
/*
    x=Lx/2
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][x];

	FFT(Ly,Ey,U,-1);

	U->r=U->r*Nhx2*(double)Nhy+Ngx2*(double)Ngy*U->i;
	U->i=0.0;
	for (ey=Ey+1, gy=Ngy1, hy=Nhy1, u=U+1, v=U+Ly1;
	     u<v;
	     ey++, gy=(gy+Ngy1)%Ly2, hy=(hy+Nhy1)%Ly2, u++, v--) {
	    t=ey->r/ey->i;
	    r=Ey[hy].r;
	    i=Ey[hy].i; w=Nhx2*(r+i*t);
			g2=Product(u->r+v->r,u->i-v->i,w*r, w*i);
	    r=Ey[gy].r;
	    i=Ey[gy].i; w=Ngx2*(r+i*t);
			h2=Product(w*r,-w*i,u->i+v->i,u->r-v->r);

	    v->r=u->r=(g2.r+h2.r)/2.0;
		 u->i=(g2.i+h2.i)/2.0; v->i=(-u->i);
	}
	u->r=u->r*Nhx2*Nhy2+Ngx2*Ngy2*u->i;
	u->i=0.0;

	FFT(Ly,Ey,U, 1);

	for (u=U, y=0; y<Ly; y++) Q[y][x]=(*(u++));
/*
    ----
*/
	for (Py=P, y=Dy1; y<=Dy2; y++) {
	    Qx=Q[(y<0)?Ly+y:y];

	    FFT(Lx,Ex,Qx, 1);

	    for (Px=(*(Py++)), x=Dx1; x<=Dx2; x++) *(Px++)=Qx[(x<0)?Lx+x:x].r;
	}
	for (Qy=Q, Gy=G+yg, Hy=H+yh, y=0; y<Ly; y++, Qy++) {
	    if (y<Ngy)
		for (Qx=(*Qy), Gx=(*(Gy++))+xg, x=0; x<Lx; x++, Qx++)
		    Qx->r=(x<Ngx)?Bg+Sg*(double)*(Gx++):0.0;
	    else
		for (Qx=(*Qy), x=0; x<Lx; x++, Qx++) Qx->r=0.0;

	    if (y<Nhy)
		for (Qx=(*Qy), Hx=(*(Hy++))+xh, x=0; x<Lx; x++, Qx++)
		    Qx->i=(x<Nhx)?Bh+Sh*(double)*(Hx++):0.0;
	    else
		for (Qx=(*Qy), x=0; x<Lx; x++, Qx++) Qx->i=0.0;

	    if (y<Ngy || y<Nhy) FFT(Lx,Ex,*Qy,-1);
	}
/*
    x=0
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][0];

	FFT(Ly,Ey,U,-1);

	U->r*=U->i; U->i=0.0;
	for (u=U+1, v=U+Ly1; u<v; u++, v--) {
	    *u=Product((u->r+v->r)/2.0,(u->i-v->i)/2.0,
		       (u->i+v->i)/2.0,(u->r-v->r)/2.0); v->r=  u->r ;
							 v->i=(-u->i);
	}
	u->r*=u->i; u->i=0.0;

	FFT(Ly,Ey,U, 1);

	for (u=U, y=0; y<Ly; y++) Q[y][0]=(*(u++));
/*
    x=1~Lx/2-1
*/
	for (x=1, Lxx=Lx1; x<Lxx; x++, Lxx--) {
	    for (u=U, v=V, y=0; y<Ly; y++) {
		*(u++)=Q[y][  x];
		*(v++)=Q[y][Lxx];
	    }
	    FFT(Ly,Ey,U,-1);
	    FFT(Ly,Ey,V,-1);

	    *U=Product((U->r+V->r)/2.0,(U->i-V->i)/2.0,
		       (U->i+V->i)/2.0,(U->r-V->r)/2.0); V->r=  U->r ;
							 V->i=(-U->i);
	    for (u=U+1, v=V+Ly1, y=1; y<Ly; y++, u++, v--) {
		*u=Product((u->r+v->r)/2.0,(u->i-v->i)/2.0,
			   (u->i+v->i)/2.0,(u->r-v->r)/2.0); v->r=  u->r ;
							     v->i=(-u->i);
	    }
	    FFT(Ly,Ey,U, 1);
	    FFT(Ly,Ey,V, 1);

	    for (u=U, v=V, y=0; y<Ly; y++) {
		Q[y][  x]=(*(u++));
		Q[y][Lxx]=(*(v++));
	    }
	}
/*
    x=Lx/2
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][x];

	FFT(Ly,Ey,U,-1);

	U->r*=U->i; U->i=0.0;
	for (u=U+1, v=U+Ly1; u<v; u++, v--) {
	    *u=Product((u->r+v->r)/2.0,(u->i-v->i)/2.0,
		       (u->i+v->i)/2.0,(u->r-v->r)/2.0); v->r=  u->r ;
							 v->i=(-u->i);
	}
	u->r*=u->i; u->i=0.0;

	FFT(Ly,Ey,U, 1);

	for (u=U, y=0; y<Ly; y++) Q[y][x]=(*(u++));
/*
    ----
*/
	Nghx=Ngx-Nhx;
	Nghy=Ngy-Nhy;
	L=(double)Lx*(double)Ly;

	for (Py=P, y=Dy1; y<=Dy2; y++, Py++) {
	    if (y<0) {
		Qx=Q[Ly+y]; LNy=L*(double)((y>=Nghy)?Ngy:Nhy+y);
	    }
	    else {
		Qx=Q[y]; LNy=L*(double)((y>=Nghy)?Ngy-y:Nhy);
	    }
	    FFT(Lx,Ex,Qx, 1);

	    for (Px=(*Py), x=Dx1; x<=Dx2; x++, Px++) {
		if (x<0)
		    msd=(*Px-2.0*Qx[Lx+x].r)/(double)((x>=Nghx)?Ngx:Nhx+x);
		else
		    msd=(*Px-2.0*Qx[x].r)/(double)((x>=Nghx)?Ngx-x:Nhx);

		(void)printf("%d\t%d\t%lf\n",
			     x+Dx0,y+Dy0,sqrt(((msd<0.0)?0.0:msd)/LNy));
	    }
	}
	return 0;
}
